import functools as ft
from typing import Mapping

from pydeepmerge import deep_merge
from pydeepmerge._deep_merge import MergeStrategy
from pydeepmerge.types import Key


def recurse(strategy: MergeStrategy) -> MergeStrategy:
    @ft.wraps(strategy)
    def newstrategy(left, right):
        if isinstance(left, Mapping) and isinstance(right, Mapping):
            return deep_merge(left, right, merge_strategy=newstrategy)
        return strategy(left, right)
    return newstrategy


def keep_default(stategy: MergeStrategy) -> MergeStrategy:
    @ft.wraps(stategy)
    def newstrategy(left, right):
        if left is Key.KeyNotFound:
            return right
        return stategy(left, right)
    return newstrategy


def disallow_mixed(strategy: MergeStrategy) -> MergeStrategy:
    @ft.wraps(strategy)
    @keep_default
    def newstrategy(left, right):
        if isinstance(left, Mapping) ^ isinstance(right, Mapping):
            raise TypeError("Combining a mapping with a non-mapping disallowed")
        return strategy(left, right)
    return newstrategy



'''
Merge strategies for the deep_merge function
'''

from pydeepmerge.strategies._strategies import (  # noqa: F401
    meta_operator_merge,
)

# The merge strategy prefer_right is still a part of this package's
# namespace, but is added dynamically at the top level package's init file
# to avoid some errors with circular imports



from collections.abc import Mapping
from pydeepmerge.types import Key
from typing import (
    Any,
    Callable,
    Dict,
    List
)
import logging

MergeStrategy = Callable[[Any, Any], Any]

logging.basicConfig(level=logging.INFO)


def prefer_right(left_value: Any, right_value: Any) -> Any:
    '''
    A merge strategy for prefering the right-hand value when
    merging Mappings while also recursing if the right-hand value
    is itself a Mapping
    '''
    logging.debug(
        "prefer_right: left_value: %s, right_value: %s",
        left_value,
        right_value
    )
    if left_value is Key.KeyNotFound:
        return right_value

    if isinstance(right_value, Mapping):
        if not isinstance(left_value, Mapping):
            left_value = {}
        return deep_merge(left_value, right_value, merge_strategy=prefer_right)

    return right_value


def deep_merge(
    *mappings: List[Mapping],
    merge_strategy: MergeStrategy = prefer_right
) -> Dict:
    '''
    Accepts a series of mappings and optionally a merge strategy
    and produces a dictionary from applying the merge strategy to the values

    The merge_strategy should be a function that accepts two values and returns
    whatever the value of the merge between those two values are. If the current
    dictionary does not have a key for the value in the merging dictionary,
    the input value of the left_value function will be Key.KeyNotFound.

    This type was used instead of None to allow for None to still be a
    valid value in a dictionary.
    '''
    result = {}
    for mapping in mappings:
        for key, right_value in mapping.items():
            logging.debug("deep_merge: key: %s", key)
            left_value = result.get(key, Key.KeyNotFound)
            result[key] = merge_strategy(left_value, right_value)
            logging.debug("result_dict: %s", result)
    return result



from yaml import load
try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader


def yaml_loader(filename):
    with open(filename, 'r') as f:
        data = load(f, Loader=Loader)

    return data



'''
A lightweight package to perform deep merges of Mapping objects
'''

import pydeepmerge._deep_merge as _deep_merge

from pydeepmerge._deep_merge import deep_merge  # noqa: F401

# Note that if `strategies` is imported before
# `deep_merge`, the code imports of `deep_merge`
# in `strategies` will fail.
import pydeepmerge.strategies as strategies

# Since prefer_right is a merge strategy, it should be
# in the strategies subpackage, but it's also the default
# merge strategy, and recursive, so if it is not
# in the _deep_merge module it'll cause a circular
# import
strategies.prefer_right = _deep_merge.prefer_right

from pydeepmerge._merge_configs import merge_configs  # noqa: F401, E402



from typing import Callable, Mapping, Union

from pydeepmerge._deep_merge import MergeStrategy
from pydeepmerge.strategies.decorators import (
    recurse,
    keep_default,
    disallow_mixed,
)


def meta_operator_merge(
    op: MergeStrategy,
    mixed: Union[bool, MergeStrategy] = False
) -> MergeStrategy:
    """
    Converts a binary operator `op` such as those from the
    `operator` package to a `MergeStrategy`.

    If `mixed` is `False`, and exactly one of `left` or
    `right` are `Mapping`s during a merge, then a
    `TypeError` is raised (this is the default).

    Therefore, if `mixed` is set to `True`, then the
    provided operator `op` must handle the case where
    exactly one of `left` or `right` is a mapping.

    Otherwise, `mixed` must be a `MergeStrategy` (a
    `Callable`) which handles this case.
    """
    @recurse
    @keep_default
    def operator_merge(left, right):
        if (isinstance(mixed, Callable)
            and (isinstance(left, Mapping) ^ isinstance(right, Mapping))
        ):  # noqa: E124
            return mixed(left, right)
        return op(left, right)
    if mixed:
        return operator_merge
    return disallow_mixed(operator_merge)



import pkg_resources
import logging


def load_parsers():
    parsers = {}

    for entry_point in pkg_resources.iter_entry_points('pydeepmerge.config_parsers'):
        logging.debug('Found entry_point for %s', entry_point.name)
        try:
            entry_point.require()
            parsers[entry_point.name] = entry_point.load()
        except pkg_resources.DistributionNotFound:
            pass

    return parsers



'''
A module with errors used for PyDeepMerge
'''


class DeepMergeError(Exception):
    'An abstract base class for errors produced by this package'


class ParserNotAvailableError(DeepMergeError):
    '''
    An error raised when a file does not have
    an appropriate parser to read it currently installed
    '''


class FileTypeError(DeepMergeError):
    '''
    Error raised when file has no file type extension
    '''



from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    long_description = f.read()

with open('VERSION', 'r') as f:
    version = f.read()

description = \
    'A lightweight library to perform deep merges of python dictionaries'

setup(
    name='pydeepmerge',
    version=version,
    license='MIT',
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/taliamax/pydeepmerge',
    author='Natalia Maximo',
    author_email='iam@natalia.dev',
    python_requires=">=3.5",
    packages=find_packages(),
    keywords='mapping dictionary library merge',
    extras_require={
        'yaml': ['pyyaml'],
        'test': ['pytest', 'pytest-cov', 'coveralls', 'flake8']
    },
    entry_points={
        'pydeepmerge.config_parsers': [
            '.json = pydeepmerge.parsers.json:json_parser',
            '.yaml = pydeepmerge.parsers.yaml:yaml_loader [yaml]',
            '.yml = pydeepmerge.parsers.yaml:yaml_loader [yaml]'
        ]
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



from pathlib import Path

from pydeepmerge import deep_merge
from pydeepmerge.errors import (
    FileTypeError,
    ParserNotAvailableError
)
from pydeepmerge.parsers import load_parsers
from pydeepmerge.strategies import prefer_right
from pydeepmerge.types import MergeStrategy
from typing import (
    Dict,
    List,
)


def merge_configs(
    *files: List[str],
    merge_strategy: MergeStrategy = prefer_right,
    auto: bool = True,
    strict: bool = True,
    extra_parsers: Dict = {}
) -> Dict:
    '''
    A function that takes a sequence of config filenames and produces
    a dictionary resulting from all the files being processed and then merged
    with a specified merge strategy.

    To use custom parsers, pass a flat dictionary of filetypes and parsers.
    If `auto` is set to False, use only the parsers explicitly passed
    through the `extra_parsers` parameter.

    If `strict` is set to True, all file names must have the same extension.
    This also applies to file type variations like `.yml` vs `.yaml`
    '''

    if strict:
        file_iter = iter(files)
        first_file_name = next(file_iter)

        file_type = Path(first_file_name).suffix

        if file_type == '':
            raise FileTypeError(
                'File "%s" has no file type extension' % first_file_name
            )

        for file_name in file_iter:
            if Path(file_name).suffix != file_type:
                raise FileTypeError(
                    'File "%s" has improper filetype. '
                    'Expected file type of "%s"' % (file_name, file_type)
                )

    parsers = {}
    if auto:
        parsers = load_parsers()

    parsers.update(extra_parsers)

    file_data = []

    for file_name in files:
        file_type = Path(file_name).suffix

        if file_type not in parsers:
            raise ParserNotAvailableError('No parser for "%s"' % file_type)

        parser = parsers[file_type]
        file_data.append(parser(file_name))

    return deep_merge(*file_data, merge_strategy=merge_strategy)
